#! /usr/bin/perl -w

# Copyright 2004 Jon Griffiths (jon_p_griffiths@yahoo.com)
#
# This script may be distributed, modifed and used without restriction provided
# that the above copyright notice is not removed.
#
# Purpose:
# This script checks that each include file used by the source files in a dll
# is actually needed by that source file. An include is needed if:
#  A) Removing it causes the build to fail
#  B) Removing it causes new warnings when rebuilding
#  C) Removing it from the source file does not change the build dependencies
# 
# To ensure we don't remove a file that is used only for prototype validation,
# we change the build flags in the makefile to give warnings if a function
# is declared without a prototype, and rely on check B) above.
#
# Check C) is really only needed while we have non-standard includes in the
# includes/ directory. If a standard include brings in a file, it is pointless
# to include that file in the source file again (it just slows down the build).
#
# Usage:
# Before running this script, run './configure && make depend && make' from
# the source directory. This script needs the dependency information to run,
# and all supplimental libraries for each dll (libuuid/libport etc) should be
# built before running.
#
# Run this script from a directory in the wine source tree that contains a
# Makefile. It will determine the files to work on automatically.
#
# The script modifies the files in place. Expect it to take 4-5 times as long
# as a regular Wine build to run.
#
# Notes:
# To be absolutely sure all uneeded includes have been eliminated, you should
# run this script until it produces no changes. This is because two or more
# uneeded includes may cause a dependency to exist on the same header.
# Removing one at a time will not cause that dependency to change, however if
# one is removed the other(s) may be able to be removed also.

use strict;
use bytes;

my @files;
my %dependencies;
my @lines;

sub modify_makefile()
{
  my @mlines = ();

  open(MY_FILE, "<Makefile") || die;
  while(<MY_FILE>)
  {
    my $mline = $_;
    # Remove debug/optimising flags (speeds up the build)
    $mline =~ s/\-g \-O2//;
    $mline =~ s/\-gstabs\+//g;
    # Add warnings for functions without definitions. Then if we remove a
    # header that declares a function in a .c file, we will get a new warning
    # (and will undo the change).
    if ( $mline !~ /missing-prototypes/ )
    {
      $mline =~ s/\-Wpointer-arith/\-Wpointer-arith \-Wmissing-prototypes \-Wmissing-declarations/;
    }
    push @mlines, $mline;
  }
  close (MY_FILE);

  open(MY_FILE, ">Makefile") || die;
  foreach (@mlines)
  {
    print MY_FILE $_;
  }
  close (MY_FILE);
}
  
# Read Makefile, put dependencies into a hash on source file name
sub read_makefile_deps(\%)
{
  my $at_dependencies = 0;

  my $dep_cmd = "make depend";
  $dep_cmd = `$dep_cmd`;

  open(MY_FILE, "<Makefile") || die;
  my $filename;
  while(<MY_FILE>)
  {
    my $line = $_;
    $line =~ s/\r//g;
    $line =~ s/\n//g;
    if ( $line =~ /Dependencies\:/ )
    {
      $at_dependencies = 1;
    }
    elsif ( $at_dependencies != 0)
    {
      if ($line =~ /^([a-zA-Z0-9_\.\/]+)\.o\:(.*)/ )
      {
        $filename = $1.".c";
        $_[0]{$filename} = $2;
      }
      else
      {
        # Create a space seperated list of the dependencies for
	# this source file
        $_[0]{$filename} = $_[0]{$filename}.$line;
        $_[0]{$filename} =~ s/\s+\\/ /g;
        $_[0]{$filename} =~ s/([a-zA-Z0-9_\.\/]+)\.c/ /g;
        $_[0]{$filename} =~ s/\s+/ /g;
        $_[0]{$filename} =~ s/^ //;
      }
    }
  }
  close (MY_FILE);
}

# Read in a .c file, put lines into a list
sub read_c_file($\@)
{
  my $c_file = shift;
  my $lines = shift;

  open(MY_FILE, "<$c_file") || return 0;
  while(<MY_FILE>)
  {
    my $line = $_;
    $line =~ s/\r//g;
    $line =~ s/\n//g;
    push @{$lines}, $line;
  }
  close (MY_FILE);
  return 1;
}

sub write_c_file($\@)
{
  my $c_file = shift;
  my $lines = shift;

  open(MY_FILE, ">$c_file") || die;
  foreach (@{$lines})
  {
    print MY_FILE $_."\n";
  }
  close (MY_FILE);
}

sub compile_c_file($$)
{
  my $c_file = shift;
  my $output_file = shift;

  my $command = "touch $c_file";
  $command = `$command`;

  $output_file =~ s/\.c/\.o/g;

  # Do a full 'make'. That way if we took out an include with a #define xxx()
  # we need, we will get a link error for the non-existing function.
  $command = "make >$output_file 2>&1";
  $command = `$command`;
  return $? >> 8;
}

sub make_files()
{
  foreach my $c_file (keys %dependencies)
  {
    my @lines = ();
    next if read_c_file($c_file, @lines) == 0;

    # Get a sorted list of the original dependencies
    my @_c_deps = sort split ( / / ,$dependencies{$c_file});
    my $c_deps = join (' ', @_c_deps);

    # Write the unmodified file, get compile results to output file
    write_c_file($c_file, @lines);
    my $orig_output_file = $c_file.".orig";
    compile_c_file($c_file, $orig_output_file);

    # Comment out the includes one by one
    while (1)
    {
      my $i = 0;

      foreach (@lines)
      {
        if ( /^\s*#\s*include\s+\"/ && ! /\/\/ DONE/ &&
             ! /debug\// && ! /config/ && ! /port/ && ! /windef/ &&
             ! /test\// && ! /pshpack/ && ! /poppack/)
        {
          # Lets try getting rid of this include
          $lines[$i] = "// ".$lines[$i]."// DONE";
          last;
        }
        $i++;
      }
      last if ( $i >= $#lines );

      my $include_name = "";
      if ($lines[$i] =~ /include\s+\"([a-zA-Z0-9_\.\/]+\.h)\"/ )
      {
        $include_name = $1;
      }
      print "$c_file:$include_name:";

      write_c_file($c_file, @lines);
      my $output_file = $c_file.".new";
      my $compile_result = compile_c_file($c_file, $output_file);

      if ($compile_result != 0)
      {
        # Failed, better return the line to normal.
        # Leave the DONE indicator so we don't do it again
        $lines[$i] =~ s/^\/\/ //;
        print "Needed (failed build)\n";
      }
      else
      {
        # Check if the warnings changed
        my $diff_cmd = "diff --brief $orig_output_file $output_file";
        $diff_cmd =~ s/\.c/\.o/g;
        $diff_cmd = `$diff_cmd`;

        if (($? >> 8) != 0)
        {
          # Got some warnings we didn't have before,
          # better return the line to normal.
          # Leave the DONE indicator so we don't do it again
          $lines[$i] =~ s/^\/\/ //;
          print "Needed (new warnings)\n";
        }
        else
        {
          # Check if the dependencies changed
          my %new_dependencies;

          read_makefile_deps(%new_dependencies);
          my @_new_c_deps = sort split ( / / ,$new_dependencies{$c_file});
          my $new_c_deps = join (' ', @_new_c_deps);

          if ($c_deps eq $new_c_deps)
          {
            # Dependencies didn't change
            # better return the line to normal.
            # Leave the DONE indicator so we don't do it again
            $lines[$i] =~ s/^\/\/ //;
            print "Not needed (but no dependency change - skipping)\n";
          }
         else
         {
            print "Not needed\n";
         }
        }
      }
    }

    my $clean_command = "rm -f *.orig *.new >/dev/null 2>&1";
    $clean_command = `$clean_command`;

    # Remove any DONE lines and write the file
    open(MY_FILE, ">$c_file") || die;
    foreach (@lines)
    {
      my $out_line = $_;
      $out_line =~ s/\/\/ DONE//;
      if (! /^\/\// )
      {
        print MY_FILE $out_line."\n";
      }
    }
    close (MY_FILE);

    # Ensure that we don't get the build results of this file
    # included in the original output of the next to be modified.
    my $cleanup_cmd = `make >/dev/null 2>&1`;
  }
}

# Starts here
my $cmd = `make`;

if (($? >> 8) == 0)
{
  # Initial make OK, go through the files and check includes
  $cmd = `cp -a Makefile Makefile.save`;
  modify_makefile();
  read_makefile_deps(%dependencies);
  make_files();
  $cmd = `mv -f Makefile.save Makefile`;
}

